# The forced depolarization algorithm implemented in Genice.
import numpy as np
import random
import heapq


def shortest_paths(G, start, ends):
    """
    Find all shortests paths from the start to one of the ends.

    Returns:
    list of shortest paths from the start to one of the ends.
    """
    q = [(0, [start, ])]  # Heap of (cost, path)
    visited = set()       # Visited vertices.
    cheapest = 999999
    while len(q):
        # logger.debug(q)
        (cost, path) = heapq.heappop(q)
        if cost > cheapest:
            break
        v0 = path[-1]
        if v0 in ends:
            cheapest = cost  # first arrived
            yield path
        else:
            if v0 in visited:
                continue
            visited.add(v0)
            for v1 in G[v0]:
                if v1 not in visited:
                    heapq.heappush(q, (cost + 1, path + [v1]))


def nearest_to_origin(vertices, coord, cell):
    """
    vertices: a subset of vertices
    coord: fractional coordinates of all vertices
    cell: cell shape
    """
    v = coord[vertices]  # fancy indexing
    v -= np.floor(v + 0.5)
    d = np.sum(v * v, axis=1)
    nearest = np.argmin(d)
    # print(nearest, d[nearest])
    return vertices[nearest]


def net_polarization(dipoles):
    net = 0
    for edge in dipoles:
        net += dipoles[edge]
    return net


def cycle_dipole(cycle, dipoles):
    net = 0
    for i in range(len(cycle)):
        a, b = cycle[i - 1], cycle[i]
        net += dipoles[a, b]
    return net


def invert(cycle, G, dipoles):
    # print(cycle)
    for i in range(len(cycle)):
        a, b = cycle[i - 1], cycle[i]
        dipoles[b, a] = -dipoles[a, b]
        del dipoles[a, b]
        G.remove_edge(a, b)
        G.add_edge(b, a)


def force_depolarize(G, rpos, cell, dipoles):
    net = net_polarization(dipoles)
#     print(net)

    # all vertices.
    # It should be divided into subcells if the system is huge.
    vertices = [x for x in range(rpos.shape[0])]
    for axis in (0, 1, 2):
        unitvec = np.zeros(3)
        unitvec[axis] += 1
        while abs(net[axis]) > 0.1:
            # depolarize in z axis

            # choose a vertex
            center = random.sample(vertices, 1)[0]

            # find the apsis
            apsis = nearest_to_origin(
                vertices, rpos - (rpos[center] + unitvec / 2), cell)
            # for v in (center, apsis):
            #     print(v, rpos[v])

            # find paths and make a directed cycle
            for c2a in shortest_paths(G, center, [apsis]):
                for a2c in shortest_paths(G, apsis, [center]):
                    cycle = c2a[:-1] + a2c[:-1]
                    dip = cycle_dipole(cycle, dipoles)
                    # print(dip)
                    newnet = net - dip * 2
                    if net @ net > newnet @ newnet:
                        # accept and invert.
                        #                         print(f"axis {axis} accepted {len(cycle)}")
                        invert(cycle, G, dipoles)
                        net = newnet
                    break
                break



import random

import networkx as nx
import numpy as np

import tilecycles.Dipole as dp


def find_cycle(g, chain, order):
    # head of the snake
    head = chain[-1]
    # neck is the vertex next to head
    neck = -1
    if len(chain) > 1:
        neck = chain[-2]
    while True:
        # candidates for the next step
        candids = [i for i in g[head] if i != neck]
        # go ahead
        neck = head
        head = random.choice(candids)
        # lookup the new head in the markers
        i = order[head]
        # if it is marked as the tail end of the snake,
        if i == 0:
            # the random walk returns to the first node of the chain;
            # unmark the head.
            order[head] = -1
            # return an empty chain and a cycle
            return [], chain
        # if the random walk crosses at an intermediate node of the chain,
        elif i > 0:
            # return a chain and a cycle
            return chain[:i + 1], chain[i:]
        # otherwise, mark the new head position
        order[head] = len(chain)
        chain.append(head)


def remove_cycle(g, cycle, order):
    # unmark vertices in the cycle (except for the first vertex)
    for i in range(1, len(cycle)):
        order[cycle[i]] = -1
    # remove edges of the cycle
    for i in range(len(cycle)):
        a = cycle[i - 1]
        b = cycle[i]
        g.remove_edge(a, b)
    # remove edgeless vertices in the graph
    for a in cycle:
        if g.degree(a) == 0:
            g.remove_node(a)


def tileByCycles(g):
    # random walk path
    chain = []
    Nnode = g.number_of_nodes()
    # markers that indicate the orders in the path
    order = -np.ones(Nnode, dtype=int)
    while g.number_of_nodes() > 0:
        # if the chain is empty
        if len(chain) == 0:
            # randomly select the "head" node.
            head = random.choice(list(g.nodes()))
            chain = [head]
            # mark it as the first node.
            order[head] = 0
        # walk randomly to find a cycle.
        chain, cycle = find_cycle(g, chain, order)
        # found.
        yield cycle
        # remove it from the graph and unmark.
        remove_cycle(g, cycle, order)


def tileByCycles2(g):
    """
    Trial for more homogeneous sampling by node division.
    According to the idea by Prof. Sakuma at Akita University.
    However, it results in very huge cycles, that are not suitable for depolarization.
    Also, it is a little bit slower than tileByCycles().
    """
    def labels(nodes):
        newn = []
        for node in nodes:
            if node < 0:
                newn.append(-1-node)
            else:
                newn.append(node)
        return newn

    # Double the nodes
    # Give negative numbers for the duplicated nodes
    Nnode = g.number_of_nodes()
    for i in range(Nnode):
        doppel = -i-1
        assert not g.has_node(doppel)
        g.add_node(doppel)
        nei = [j for j in g.neighbors(i)]
        assert len(nei) == 4, g[i]
        for v in random.sample(nei, 2):
            g.remove_edge(i, v)
            g.add_edge(doppel, v)
    while g.number_of_nodes() > 0:
        cycle = []
        head = list(g.nodes())[0]
        nei = [j for j in g.neighbors(head)]
        # print(nei)
        cycle.append(head)
        last = head
        succ = random.choice(nei)
        # print(succ)
        while succ != head:
            cycle.append(succ)
            # print(cycle)
            nei = [j for j in g.neighbors(succ)]
            assert len(nei) == 2
            s2 = nei[0]
            if s2 == last:
                s2 = nei[1]
            last, succ = succ, s2
        yield labels(cycle)
        for v in cycle:
            g.remove_node(v)



def tile(pairs, Nnode):
    """
    Nnode and seed are dummpy parameters
    for the compatibility with c++ codes.
    """
    g = nx.Graph()
    g.add_edges_from(pairs)
    return [cycle for cycle in tileByCycles(g)]


def depolarize(cycles, dipoles, dd, rpos, cell):
    # pick up the traversing cycles
    cid = []
    cycdip = []
    direction = []
    for i, cycle in enumerate(cycles):
        dip = dp.cycle_dipole(cycle, dipoles)
        if not np.allclose(dip, 0):
            cid.append(i)
            cycdip.append(dip)
            direction.append(1)
    cycdip = np.array(cycdip)
    direction = np.array(direction)
    Ncycle = cycdip.shape[0]

    # reduce polarization by flipping cycles randomly
    net = dp.net_polarization(dipoles)
    count = 100
    while not np.allclose(net, 0):
        r = random.randint(0, Ncycle - 1)
        newnet = net - 2 * direction[r] * cycdip[r]
        if net @ net > newnet @ newnet:
            direction[r] = -direction[r]
            net = newnet
        count -= 1
        if count == 0:
            print("give up depolarizing completely.")
            break

    # reflect the inversions to the dipoles and graph
    for i in range(Ncycle):
        if direction[i] < 0:
            dp.invert(cycles[cid[i]], dd, dipoles)

    if not np.allclose(net, 0):
        print("Original GenIce algorithm.")
        dp.force_depolarize(dd, rpos, cell, dipoles)
        return 1
    else:
        return 0


# For odd-odd chains
def remove_path(G, path):
    for i in range(len(path) - 1):
        G.remove_edge(path[i], path[i + 1])
    for node in path:
        if len(G[node]) == 0:
            G.remove_node(node)


def odd_chains(g):
    # make a list of odd vertices
    odds = set([node for node in g if len(g[node]) % 2 == 1])

    # couple them.
    while len(odds) > 0:
        node = odds.pop()
        for path in dp.shortest_paths(g, node, odds):
            end = path[-1]
            assert end in odds
            odds.remove(end)
            yield path
            break



import random


def six(d, Nnode, MaxSize=1e99):
    """
    find a cyclic path in the given digraph.
    d: digraph (networkx.DiGraph)
    """
    head = random.randint(0, Nnode - 1)
    path = [head]
    while True:
        nexts = list(d.successors(head))
        next = random.choice(nexts)
        if next in path:
            i = path.index(next)
            if len(path[i:]) >= MaxSize:
                # print("Drop.")
                return []
            return path[i:]
        path.append(next)
        head = next


def invertCycle(d, cycle, g_footprint):
    for i in range(len(cycle)):
        a, b = cycle[i - 1], cycle[i]
        d.remove_edge(a, b)
        d.add_edge(b, a)
        # footprint
        if g_footprint.has_edge(a, b):
            g_footprint.remove_edge(a, b)



import random


def migrate(d, excess):
    """
    excess is a set of nodes having more than two outgoing edges
    """
    while len(excess) > 0:
        # random choice from candidates;
        # there may be faster ways
        head = random.choice(list(excess))
        nexts = list(d.successors(head))
        next = random.choice(nexts)
        d.remove_edge(head, next)
        d.add_edge(next, head)
        # if head's outgoing order was three
        if len(nexts) == 3:
            # remove from the excess list
            excess.remove(head)
        # if next's outgoing orderis three
        if d.out_degree(next) == 3:
            # it becomes a new excess
            excess.add(next)



#!/usr/bin/env python

# from distutils.core import setup, Extension
import codecs
# from numpy.distutils.misc_util import get_numpy_include_dirs
# import numpy
import os
import re

from setuptools import Extension, dist, find_packages, setup

# Copied from wheel package
here = os.path.abspath(os.path.dirname(__file__))

__version__ = "0.1.5.5"

# bootstrap numpy
dist.Distribution().fetch_build_eggs(['numpy'])

setup(
# ext_modules=[Extension("tilecycles_c", ["c_tilecycles.cpp", "tilecycles_c.cpp"],
#                              extra_compile_args=["-std=c++11", ],
#                              include_dirs=get_numpy_include_dirs())],
#       headers=["tilecycles_c.hpp"],
      # include_dirs=get_numpy_include_dirs(),
      name='TileCycles',
      version=__version__,
      zip_safe=False,
      # py_modules=['tilecycles'],
      description='Bla bla.',
      #long_description=README + '\n\n' +  CHANGES,
      classifiers=[
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.5",
],
    packages=find_packages(),

    author='Masakazu Matsumoto',
    author_email='vitroid@gmail.com',
    url='https://github.com/vitroid/TileCycles/',
    keywords=['tile by cycles', ],
    license='MIT',
    install_requires=['numpy', 'networkx', 'cycless'],
)
